--[[ getOrDefault
function getOrDefault(list, key, default)
  return (list or { })[key] or default
end
]]

--[[ getOrNil
function getOrNil(list, key)
  return (list or { })[key]
end
]]

-- replicate :: int -> a -> { a }
function replicate(n, x)
	local output = { }
	for i=1, n do table.insert(output, x) end
	return output
end

-- forEach :: (a -> ()) -> { a } -> ()
function forEach(a, xs)
	for _,x in ipairs(xs) do
		a(x)
	end
end

--[[ forEachKey
function forEachKey(a, xs)
	for k,_ in pairs(xs) do
		a(k)
	end
end
]]

--[[ forEachKeyValuePair
function forEachKeyValuePair(a, xs)
	for k,v in pairs(xs) do
		a(k,v)
	end
end
]]

-- keys :: { k -> _ } -> { k }
function keys(xs)
  local output = { }
  for k,_ in pairs(xs) do
    table.insert(output, k)
  end
  return output
end

--[[ values
function values(xs)
  local output = { }
  for _,v in pairs(xs) do
    table.insert(output, v)
  end
  return output
end
]]

--[[ keyValuePairs
function keyValuePairs(xs)
  local output = { }
  for k,v in pairs(xs) do
    table.insert(output, KeyValuePair(k,v))
  end
  return output
end
]]

-- Preserves order
-- map :: (a -> b) -> { k -> a } -> { k -> b }
function map(f, xs)
	local output = { }
	for k,v in pairs(xs) do
		output[k] = f(v)
	end
	return output
end

--[[ mapKeys
function mapKeys(f, xs)
  local output = { }
  for k,_ in pairs(xs) do
    table.insert(output, f(k))
  end
  return output
end
]]

--[[ mapKeyValuePairs
function mapKeyValuePairs(f, xs)
  local output = { }
  for k,v in pairs(xs) do
    table.insert(output, f(KeyValuePair(k,v)))
  end
  return output
end
]]

-- filter :: (a -> bool) -> { a } -> { b }
function filter(p, xs)
	local output = { }
	for _,v in pairs(xs) do
		if p(v) then 
			table.insert(output, v)
		end
	end
	return output
end

-- filterKeysByValue :: (a -> bool) -> { k -> a } -> { k }
function filterKeysByValue(p, xs)
  local output = { }
  for k,v in pairs(xs) do
    if p(v) then 
      table.insert(output, k)
    end
  end
  return output
end

-- foldr :: (a -> b -> b) -> b -> { a } -> b
function foldr(op, val, xs)
	for _,v in pairs(xs) do
		val = op(val, v)
	end
	return val
end

--[[ foldrByKey
function foldrByKey(op, val, xs)
  for k,_ in pairs(xs) do
    val = op(val, k)
  end
  return val
end
]]

--[[ foldrByKeyValuePair
function foldrByKeyValuePair(op, val, xs)
  for k,v in pairs(xs) do
    val = op(val, KeyValuePair(k,v))
  end
  return val
end
]]

-- any :: (a -> bool) -> { a } -> bool
function any(p, xs)
	for _,v in pairs(xs) do
		if p(v) then
			return true
		end
	end
	return false
end

--[[ all
function all(p, xs)
	return not any(
		function(x) return not p(x) end, 
		xs)
end
]]

--[[ none
function none(p, xs)
	return not any(p, xs)
end
]]

-- contains :: a -> { a } -> bool
function contains(val, xs)
	return any(
		function(x) return x == val end, 
		xs)
end

--[[
function containsKey(key, xs)
	for k,_ in pairs(xs) do
		if k == key then return true end
	end
	return false
end
]]

-- const :: a -> (b -> a)
function const(x)
  return function(...) return x end
end

-- invokeAll :: { (() -> ()) } -> ()
function invokeAll(xs)
  forEach(function(x) x() end, xs)
end

-- count :: { a } -> int
function count(list)
	return foldr(
    function(accum, x) return accum + 1 end, 0, 
    list
  )
end

-- mergeTables :: { a } -> { a } -> { a }
--[[ Merges an arbitrary amount of arrays into one.
-- Ensures:
-- 1) ({1},{2}) = {1,2}
-- 2) ({ },{2}) = {  2}
-- 3) ({1},{ }) = {  1}
-- 4) (nil,{2}) = {  2}
-- 5) ({1},nil) = {  1} ]]
function mergeTables(...)
	local output = { }
	for _,t in pairs({...}) do
		for _,v in pairs(t or { }) do
			table.insert(output, v)
		end
	end
	return output
end

-- without :: { a } -> { a } -> { a }
function without(list, values)
  return filter(
    function(x) return not contains(x, values) end,
    list)
end

--[[ withoutV :: { a } -> { a } -> { a }
function withoutV(list, ...)
  local arg = {...}
	return filter(
		function(x) return not contains(x, arg) end,
		list)
end
]]

-- Item on p(x) = true is included
-- skipWhile :: (a -> bool) -> { a } -> { a }
function skipWhile(p, xs)
  local output = { }
  local found = false
  for _,v in ipairs(xs) do
    if p(v) then
      found = true
    end
    if found then 
      table.insert(output, v)
    end
  end
  return output
end

-- Item on p(x) = false not included.
-- takeWhile :: (a -> bool) -> { a } -> { a }
function takeWhile(p, xs)
  local output = { }
  local found = true
  for _,v in ipairs(xs) do
    if not p(v) then
      found = false
    end
    if found then
      table.insert(output, v)
    end
  end
  return output
end

--[[ flip
function flip(f)
  return function(x,y) return f(y,x) end
end
]]

-- End of File.